@@ -25,6 +25,9 @@ __copyright__ = "Copyright (c) 2011. Orca Team."
 __license__   = "LGPL"
 
 from gi.repository import GLib
+import gi
+gi.require_version('Atspi', '2.0') 
+from gi.repository import Atspi
 import pyatspi
 import queue
 import threading
@@ -61,20 +64,49 @@ class EventManager:
         self._ignoredEvents = ['object:bounds-changed',
                                'object:state-changed:defunct',
                                'object:property-change:accessible-parent']
+        self._parentsOfDefunctDescendants = []
+
+        orca_state.device = None
+        self.newKeyHandlingActive = False
+        self.legacyKeyHandlingActive = False
+        self.forceLegacyKeyHandling = False
+
         debug.println(debug.LEVEL_INFO, 'Event manager initialized', True)
 
     def activate(self):
-        """Called when this presentation manager is activated."""
+        """Called when this event manager is activated."""
 
         debug.println(debug.LEVEL_INFO, 'EVENT MANAGER: Activating', True)
-        self._registerListener("window:activate")
-        self._registerListener("window:deactivate")
-        self._registerListener("object:children-changed")
-        self._registerListener("mouse:button")
-        self.registerKeystrokeListener(self._processKeyboardEvent)
+        self.setKeyHandling(False)
+
         self._active = True
         debug.println(debug.LEVEL_INFO, 'EVENT MANAGER: Activated', True)
 
+    def activateNewKeyHandling(self):
+        if not self.newKeyHandlingActive:
+            try:
+                orca_state.device = Atspi.Device.new()
+            except:
+                self.forceLegacyKeyHandling = True
+                self.activateLegacyKeyHandling()
+                return
+            orca_state.device.event_count = 0
+            orca_state.device.key_watcher = orca_state.device.add_key_watcher(self._processNewKeyboardEvent)
+            self.newKeyHandlingActive = True
+
+    def activateLegacyKeyHandling(self):
+        if not self.legacyKeyHandlingActive:
+            self.registerKeystrokeListener(self._processKeyboardEvent)
+            self.legacyKeyHandlingActive = True
+
+    def setKeyHandling(self, new):
+        if new and not self.forceLegacyKeyHandling:
+            self.deactivateLegacyKeyHandling()
+            self.activateNewKeyHandling()
+        else:
+            self.deactivateNewKeyHandling()
+            self.activateLegacyKeyHandling()
+
     def deactivate(self):
         """Called when this event manager is deactivated."""
 
@@ -83,9 +115,19 @@ class EventManager:
         for eventType in self._scriptListenerCounts.keys():
             self.registry.deregisterEventListener(self._enqueue, eventType)
         self._scriptListenerCounts = {}
-        self.deregisterKeystrokeListener(self._processKeyboardEvent)
+        self.deactivateLegacyKeyHandling()
         debug.println(debug.LEVEL_INFO, 'EVENT MANAGER: Deactivated', True)
 
+    def deactivateNewKeyHandling(self):
+        if self.newKeyHandlingActive:
+            orca_state.device = None
+            self.newKeyHandlingActive = False;
+
+    def deactivateLegacyKeyHandling(self):
+        if self.legacyKeyHandlingActive:
+            self.deregisterKeystrokeListener(self._processKeyboardEvent)
+            self.legacyKeyHandlingActive = False;
+
     def ignoreEventTypes(self, eventTypeList):
         for eventType in eventTypeList:
             if not eventType in self._ignoredEvents:
@@ -99,10 +141,18 @@ class EventManager:
     def _ignore(self, event):
         """Returns True if this event should be ignored."""
 
+        anydata = event.any_data
+        if isinstance(anydata, str) and len(anydata) > 100:
+            anydata = "%s (...)" % anydata[0:100]
+
+        source = str(event.source)
+        if len(source) > 100:
+            source = "%s (...) ]" % source[0:100]
+
         debug.println(debug.LEVEL_INFO, '')
         msg = 'EVENT MANAGER: %s for %s in %s (%s, %s, %s)' % \
-              (event.type, event.source, event.host_application,
-               event.detail1,event.detail2, event.any_data)
+              (event.type, source, event.host_application,
+               event.detail1,event.detail2, anydata)
         debug.println(debug.LEVEL_INFO, msg, True)
 
         if not self._active:
@@ -120,8 +170,19 @@ class EventManager:
             debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
+        if event.type.startswith('mouse:button'):
+            msg = 'EVENT MANAGER: Not ignoring because event type is never ignored'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return False
+
+        if self._inDeluge() and self._ignoreDuringDeluge(event):
+            msg = 'EVENT MANAGER: Ignoring event type due to deluge'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         script = orca_state.activeScript
-        if event.type.startswith('object:children-changed:add'):
+        if event.type.startswith('object:children-changed') \
+           or event.type.startswith('object:state-changed:sensitive'):
             if not script:
                 msg = 'EVENT MANAGER: Ignoring because there is no active script'
                 debug.println(debug.LEVEL_INFO, msg, True)
@@ -131,21 +192,14 @@ class EventManager:
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
 
-        # This should ultimately be changed as there are valid reasons
-        # to handle these events at the application level.
-        if event.type.startswith('object:children-changed:remove') \
-           and event.source != self._desktop:
-            msg = 'EVENT MANAGER: Ignoring because event type is ignored'
-            debug.println(debug.LEVEL_INFO, msg, True)
-            return True
-
-        if event.type.startswith('object:text-changed') and event.type.endswith('system'):
+        if event.type.startswith('object:text-changed') \
+           and self.EMBEDDED_OBJECT_CHARACTER in event.any_data \
+           and not event.any_data.replace(self.EMBEDDED_OBJECT_CHARACTER, ""):
             # We should also get children-changed events telling us the same thing.
             # Getting a bunch of both can result in a flood that grinds us to a halt.
-            if event.any_data == self.EMBEDDED_OBJECT_CHARACTER:
-                msg = 'EVENT MANAGER: Ignoring because changed text is embedded object'
-                debug.println(debug.LEVEL_INFO, msg, True)
-                return True
+            msg = 'EVENT MANAGER: Ignoring because changed text is only embedded objects'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
 
         try:
             # TODO - JD: For now we won't ask for the name. Simply asking for the name should
@@ -160,75 +214,167 @@ class EventManager:
             msg = 'ERROR: Event is from potentially-defunct source'
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
+
+        if state.isEmpty():
+            msg = 'EVENT MANAGER: Ignoring event due to empty state set'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         if state.contains(pyatspi.STATE_DEFUNCT):
             msg = 'ERROR: Event is from defunct source'
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
-
-        if event.type.startswith('object:children-changed:add'):
-            if role in [pyatspi.ROLE_MENU,
-                        pyatspi.ROLE_LAYERED_PANE,
+        if event.type.startswith('object:property-change:accessible-name'):
+            if role in [pyatspi.ROLE_CANVAS,
+                        pyatspi.ROLE_ICON,
+                        pyatspi.ROLE_LABEL,      # gnome-shell spam
+                        pyatspi.ROLE_LIST_ITEM,  # Web app spam
+                        pyatspi.ROLE_LIST,       # Web app spam
+                        pyatspi.ROLE_SECTION,    # Web app spam
+                        pyatspi.ROLE_TABLE_ROW,  # Thunderbird spam
+                        pyatspi.ROLE_TABLE_CELL, # Thunderbird spam
+                        pyatspi.ROLE_MENU,
                         pyatspi.ROLE_MENU_ITEM]:
                 msg = 'EVENT MANAGER: Ignoring event type due to role'
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
-
-        if event.type.startswith('object:property-change:accessible-name'):
-            if role in [pyatspi.ROLE_CANVAS,
-                        pyatspi.ROLE_ICON]:
-                msg = 'EVENT MANAGER: Ignoring event type due to role'
+        elif event.type.startswith('object:property-change:accessible-value'):
+            if role == pyatspi.ROLE_SPLIT_PANE and not state.contains(pyatspi.STATE_FOCUSED):
+                msg = 'EVENT MANAGER: Ignoring event type due to role and state'
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
-
-        if event.type.startswith('object:state-changed:sensitive'):
+        elif event.type.startswith('object:text-changed:insert') and event.detail2 > 1000 \
+             and role in [pyatspi.ROLE_TEXT, pyatspi.ROLE_STATIC]:
+            msg = 'EVENT MANAGER: Ignoring because inserted text has more than 1000 chars'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+        elif event.type.startswith('object:state-changed:sensitive'):
             if role in [pyatspi.ROLE_MENU_ITEM,
+                        pyatspi.ROLE_MENU,
+                        pyatspi.ROLE_FILLER,
+                        pyatspi.ROLE_PANEL,
                         pyatspi.ROLE_CHECK_MENU_ITEM,
                         pyatspi.ROLE_RADIO_MENU_ITEM]:
                 msg = 'EVENT MANAGER: Ignoring event type due to role'
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
-
-        if event.type.startswith('object:state-changed:showing'):
+        elif event.type.startswith('object:state-changed:selected'):
+            if not event.detail1 and role in [pyatspi.ROLE_PUSH_BUTTON]:
+                msg = 'EVENT MANAGER: Ignoring event type due to role and detail1'
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return True
+        elif event.type.startswith('object:state-changed:showing'):
             if role not in [pyatspi.ROLE_ALERT,
                             pyatspi.ROLE_ANIMATION,
                             pyatspi.ROLE_INFO_BAR,
+                            pyatspi.ROLE_MENU,
                             pyatspi.ROLE_NOTIFICATION,
+                            pyatspi.ROLE_DIALOG,
                             pyatspi.ROLE_PANEL,
                             pyatspi.ROLE_STATUS_BAR,
                             pyatspi.ROLE_TOOL_TIP]:
                 msg = 'EVENT MANAGER: Ignoring event type due to role'
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
+            if role == pyatspi.ROLE_PANEL:
+                if not event.detail1:
+                    msg = 'EVENT MANAGER: Ignoring event type due to role and detail1'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return True
+                try:
+                    if not event.source.name:
+                        msg = 'EVENT MANAGER: Ignoring event type due to role and lack of name'
+                        debug.println(debug.LEVEL_INFO, msg, True)
+                        return True
+                except:
+                    msg = 'EVENT MANAGER: Ignoring event from dead source'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return True
 
-        if event.type.startswith('object:children-changed:add') \
+        elif event.type.startswith('object:selection-changed'):
+            if event.source in self._parentsOfDefunctDescendants:
+                msg = 'EVENT MANAGER: Ignoring event from parent of defunct descendants'
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return True
+
+            try:
+                _name = event.source.name
+            except:
+                msg = 'EVENT MANAGER: Ignoring event from dead source'
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return True
+
+        if event.type.startswith('object:children-changed') \
            or event.type.startswith('object:active-descendant-changed'):
+            if role in [pyatspi.ROLE_MENU,
+                        pyatspi.ROLE_LAYERED_PANE,
+                        pyatspi.ROLE_MENU_ITEM]:
+                msg = 'EVENT MANAGER: Ignoring event type due to role'
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return True
             if not event.any_data:
                 msg = 'ERROR: Event any_data lacks child/descendant'
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
+            if event.type.endswith('remove'):
+                if event.any_data == orca_state.locusOfFocus:
+                    msg = 'EVENT MANAGER: Locus of focus is being destroyed'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return False
+
+                try:
+                    _name = orca_state.locusOfFocus.name
+                except:
+                    msg = 'EVENT MANAGER: Locus of focus is dead.'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return False
+                else:
+                    msg = 'EVENT MANAGER: Locus of focus: %s' % orca_state.locusOfFocus
+                    debug.println(debug.LEVEL_INFO, msg, True)
+
             try:
-                state = event.any_data.getState()
-                role = event.any_data.getRole()
+                childState = event.any_data.getState()
+                childRole = event.any_data.getRole()
+                name = event.any_data.name
+                defunct = False
             except:
                 msg = 'ERROR: Event any_data contains potentially-defunct child/descendant'
                 debug.println(debug.LEVEL_INFO, msg, True)
+                defunct = True
+            else:
+                defunct = childState.contains(pyatspi.STATE_DEFUNCT)
+                if defunct:
+                    msg = 'ERROR: Event any_data contains defunct child/descendant'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+
+            if defunct:
+                if state.contains(pyatspi.STATE_MANAGES_DESCENDANTS) \
+                   and event.source not in self._parentsOfDefunctDescendants:
+                    self._parentsOfDefunctDescendants.append(event.source)
                 return True
-            if state.contains(pyatspi.STATE_DEFUNCT):
-                msg = 'ERROR: Event any_data contains defunct child/descendant'
-                debug.println(debug.LEVEL_INFO, msg, True)
-                return True
+
+            if event.source in self._parentsOfDefunctDescendants:
+                self._parentsOfDefunctDescendants.remove(event.source)
 
             # This should be safe. We do not have a reason to present a newly-added,
             # but not focused image. We do not need to update live regions for images.
             # This is very likely a completely and utterly useless event for us. The
             # reason for ignoring it here rather than quickly processing it is the
             # potential for event floods like we're seeing from matrix.org.
-            if role == pyatspi.ROLE_IMAGE:
+            if childRole == pyatspi.ROLE_IMAGE:
                 msg = 'EVENT MANAGER: Ignoring event type due to role'
                 debug.println(debug.LEVEL_INFO, msg, True)
                 return True
 
+            # In normal apps we would have caught this from the parent role.
+            # But gnome-shell has panel parents adding/removing menu items.
+            if event.type.startswith('object:children-changed'):
+                if childRole == pyatspi.ROLE_MENU_ITEM:
+                    msg = 'EVENT MANAGER: Ignoring event type due to child role'
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return True
+
         msg = 'EVENT MANAGER: Not ignoring due to lack of cause'
         debug.println(debug.LEVEL_INFO, msg, True)
         return False
@@ -260,7 +406,7 @@ class EventManager:
         if debug.debugEventQueue:
             debug.println(debug.LEVEL_ALL, "           ...released")
 
-    def _queuePrintln(self, e, isEnqueue=True):
+    def _queuePrintln(self, e, isEnqueue=True, isPrune=None):
         """Convenience method to output queue-related debugging info."""
 
         if isinstance(e, input_event.KeyboardEvent):
@@ -268,12 +414,19 @@ class EventManager:
         elif isinstance(e, input_event.BrailleEvent):
             data = "'%s'" % repr(e.event)
         elif not debug.eventDebugFilter or debug.eventDebugFilter.match(e.type):
+            anydata = e.any_data
+            if isinstance(anydata, str) and len(anydata) > 100:
+                anydata = "%s (...)" % anydata[0:100]
             data = "%s (%s,%s,%s) from %s" % \
-                   (e.source, e.detail1, e.detail2, e.any_data, e.host_application)
+                   (e.source, e.detail1, e.detail2, anydata, e.host_application)
         else:
             return
 
-        if isEnqueue:
+        if isPrune:
+            string = "EVENT MANAGER: Pruning %s %s" % (e.type, data)
+        elif isPrune is not None:
+            string = "EVENT MANAGER: Not pruning %s %s" % (e.type, data)
+        elif isEnqueue:
             string = "EVENT MANAGER: Queueing %s %s" % (e.type, data)
         else:
             string = "EVENT MANAGER: Dequeued %s %s" % (e.type, data)
@@ -310,6 +463,11 @@ class EventManager:
 
         self._queuePrintln(e)
 
+        if self._inFlood() and self._prioritizeDuringFlood(e):
+            msg = 'EVENT MANAGER: Pruning event queue due to flood.'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            self._pruneEventsDuringFlood()
+
         asyncMode = self._asyncMode
         if isObjectEvent:
             app = e.source.getApplication()
@@ -317,9 +475,15 @@ class EventManager:
                 toolkitName = app.toolkitName
             except:
                 toolkitName = None
-            if toolkitName in self._synchronousToolkits \
-               or isinstance(e, input_event.MouseButtonEvent):
+            if isinstance(e, input_event.MouseButtonEvent):
+                asyncMode = True
+            elif toolkitName in self._synchronousToolkits:
                 asyncMode = False
+            elif e.type.startswith("object:children-changed"):
+                try:
+                    asyncMode = e.source.getRole() == pyatspi.ROLE_TABLE
+                except:
+                    asyncMode = True
             script = _scriptManager.getScript(app, e.source)
             script.eventCache[e.type] = (e, time.time())
 
@@ -330,6 +494,23 @@ class EventManager:
         if debug.debugEventQueue:
             self._enqueueCount -= 1
 
+    def _isNoFocus(self):
+        if orca_state.locusOfFocus or orca_state.activeWindow or orca_state.activeScript:
+            return False
+
+        msg = 'EVENT MANAGER: No focus'
+        debug.println(debug.LEVEL_SEVERE, msg, True)
+        return True
+
+    def _onNoFocus(self):
+        if not self._isNoFocus():
+            return False
+
+        defaultScript = _scriptManager.getDefaultScript()
+        _scriptManager.setActiveScript(defaultScript, 'No focus')
+        defaultScript.idleMessage()
+        return False
+
     def _dequeue(self):
         """Handles all events destined for scripts. Called by the GTK
         idle thread."""
@@ -366,24 +547,9 @@ class EventManager:
                                   % event.type)
                 debug.objEvent = None
 
-            # [[[TODO: HACK - it would seem logical to only do this if we
-            # discover the queue is empty, but this inroduces a hang for
-            # some reason if done inside an acquire/release block for a
-            # lock.  So...we do it here.]]]
-            #
-            try:
-                noFocus = not (orca_state.activeScript or orca_state.locusOfFocus)
-            except:
-                noFocus = True
-
             self._gidleLock.acquire()
             if self._eventQueue.empty():
-                if noFocus:
-                    if self._gilSleepTime:
-                        time.sleep(self._gilSleepTime)
-                    fullMessage = messages.NO_FOCUS
-                    defaultScript = _scriptManager.getDefaultScript()
-                    defaultScript.presentMessage(fullMessage, '')
+                GLib.timeout_add(2500, self._onNoFocus)
                 self._gidleId = 0
                 rerun = False # destroy and don't call again
             self._gidleLock.release()
@@ -542,7 +708,7 @@ class EventManager:
         """Returns the script associated with event."""
 
         if event.type.startswith("mouse:"):
-            return orca_state.activeScript
+            return _scriptManager.getScriptForMouseButtonEvent(event)
 
         script = None
         app = None
@@ -604,9 +770,8 @@ class EventManager:
         
         if not script:
             script = self._getScriptForEvent(event)
-
-        if not script:
-            return False, "There is no script for this event."
+            if not script:
+                return False, "There is no script for this event."
 
         if script == orca_state.activeScript:
             return False, "The script for this event is already active."
@@ -614,9 +779,22 @@ class EventManager:
         if not script.isActivatableEvent(event):
             return False, "The script says not to activate for this event."
 
+        if script.forceScriptActivation(event):
+            return True, "The script insists it should be activated for this event."
+
         eType = event.type
+
         if eType.startswith('window:activate'):
-            return True, "window:activate event"
+            windowActivation = True
+        else:
+            windowActivation = eType.startswith('object:state-changed:active') \
+                and event.detail1 and role == pyatspi.ROLE_FRAME
+
+        if windowActivation:
+            if event.source != orca_state.activeWindow:
+                return True, "Window activation"
+            else:
+                return False, "Window activation for already-active window"
 
         if eType.startswith('focus') \
            or (eType.startswith('object:state-changed:focused')
@@ -636,6 +814,132 @@ class EventManager:
 
         return False, "No reason found to activate a different script."
 
+    def _eventSourceIsDead(self, event):
+        try:
+            name = event.source.name
+        except:
+            msg = "EVENT MANAGER: source of %s is dead" % event.type
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
+        return False
+
+    def _ignoreDuringDeluge(self, event):
+        """Returns true if this event should be ignored during a deluge."""
+
+        if self._eventSourceIsDead(event):
+            return True
+
+        ignore = ["object:text-changed:delete",
+                  "object:text-changed:insert",
+                  "object:text-changed:delete:system",
+                  "object:text-changed:insert:system",
+                  "object:text-attributes-changed",
+                  "object:children-changed:add",
+                  "object:children-changed:add:system",
+                  "object:property-change:accessible-name",
+                  "object:property-change:accessible-description",
+                  "object:state-changed:showing",
+                  "object:state-changed:sensitive"]
+
+        if event.type not in ignore:
+            return False
+
+        return event.source != orca_state.locusOfFocus
+
+    def _inDeluge(self):
+        size = self._eventQueue.qsize()
+        if size > 100:
+            msg = 'EVENT MANAGER: DELUGE! Queue size is %i' % size
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
+        return False
+
+    def _processDuringFlood(self, event):
+        """Returns true if this event should be processed during a flood."""
+
+        if self._eventSourceIsDead(event):
+            return False
+
+        ignore = ["object:text-changed:delete",
+                  "object:text-changed:insert",
+                  "object:text-changed:delete:system",
+                  "object:text-changed:insert:system",
+                  "object:text-attributes-changed",
+                  "object:children-changed:add",
+                  "object:children-changed:add:system",
+                  "object:property-change:accessible-name",
+                  "object:property-change:accessible-description",
+                  "object:state-changed:showing",
+                  "object:state-changed:sensitive"]
+
+        if event.type not in ignore:
+            return True
+
+        return event.source == orca_state.locusOfFocus
+
+    def _prioritizeDuringFlood(self, event):
+        """Returns true if this event should be prioritized during a flood."""
+
+        if event.type.startswith("object:state-changed:focused"):
+            return event.detail1
+
+        if event.type.startswith("object:state-changed:selected"):
+            return event.detail1
+
+        if event.type.startswith("object:text-selection-changed"):
+            return True
+
+        if event.type.startswith("window:activate"):
+            return True
+
+        if event.type.startswith("window:deactivate"):
+            return True
+
+        if event.type.startswith("object:state-changed:active"):
+            return event.source.getRole() in [pyatspi.ROLE_FRAME, pyatspi.ROLE_WINDOW]
+
+        if event.type.startswith("document:load-complete"):
+            return True
+
+        if event.type.startswith("object:state-changed:busy"):
+            return True
+
+        return False
+
+    def _pruneEventsDuringFlood(self):
+        """Gets rid of events we don't care about during a flood."""
+
+        oldSize = self._eventQueue.qsize()
+
+        newQueue = queue.Queue(0)
+        while not self._eventQueue.empty():
+            try:
+                event = self._eventQueue.get()
+            except Empty:
+                continue
+
+            if self._processDuringFlood(event):
+                newQueue.put(event)
+                self._queuePrintln(event, isPrune=False)
+            self._eventQueue.task_done()
+
+        self._eventQueue = newQueue
+        newSize = self._eventQueue.qsize()
+
+        msg = 'EVENT MANAGER: %i events pruned. New size: %i' % ((oldSize - newSize), newSize)
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+    def _inFlood(self):
+        size = self._eventQueue.qsize()
+        if size > 50:
+            msg = 'EVENT MANAGER: FLOOD? Queue size is %i' % size
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
+        return False
+
     def _processObjectEvent(self, event):
         """Handles all object events destined for scripts.
 
@@ -651,12 +955,7 @@ class EventManager:
                 if event.source == self._desktop:
                     _scriptManager.reclaimScripts()
                     return
-            except (LookupError, RuntimeError):
-                # If we got this error here, we'll get it again when we
-                # attempt to get the state, catch it, and clean up.
-                pass
             except:
-                debug.printException(debug.LEVEL_WARNING)
                 return
 
         if eType.startswith("window:") and not eType.endswith("create"):
@@ -671,33 +970,25 @@ class EventManager:
                 if role == pyatspi.ROLE_FRAME:
                     _scriptManager.reclaimScripts()
 
-        # Clean up any flat review context so that Orca does not get
-        # confused (see bgo#609633)
-        #
-        if eType.startswith("window:deactivate") \
-           and orca_state.activeScript \
-           and orca_state.activeScript.flatReviewContext \
-           and orca_state.activeScript.app == event.host_application:
-            orca_state.activeScript.flatReviewContext = None
-
         try:
             state = event.source.getState()
-        except (LookupError, RuntimeError):
-            msg = 'ERROR: Could not process event: %s' % eType
-            debug.println(debug.LEVEL_WARNING, msg, True)
-            if eType.startswith("window:deactivate"):
-                orca_state.locusOfFocus = None
-                orca_state.activeWindow = None
-            return
         except:
-            return
+            isDefunct = True
+            msg = 'ERROR: Exception getting state for event source'
+            debug.println(debug.LEVEL_WARNING, msg, True)
+        else:
+            isDefunct = state.contains(pyatspi.STATE_DEFUNCT)
 
-        if state and state.contains(pyatspi.STATE_DEFUNCT):
+        if isDefunct:
             msg = 'EVENT MANAGER: Ignoring defunct object: %s' % event.source
             debug.println(debug.LEVEL_INFO, msg, True)
-            if eType.startswith("window:deactivate"):
+            if eType.startswith("window:deactivate") or eType.startswith("window:destroy") \
+               and orca_state.activeWindow == event.source:
+                msg = 'EVENT MANAGER: Clearing active window, script, and locus of focus'
+                debug.println(debug.LEVEL_INFO, msg, True)
                 orca_state.locusOfFocus = None
                 orca_state.activeWindow = None
+                orca_state.activeScript = None
             return
 
         if state and state.contains(pyatspi.STATE_ICONIFIED):
@@ -705,9 +996,29 @@ class EventManager:
             debug.println(debug.LEVEL_INFO, msg, True)
             return
 
+        if self._inFlood():
+            if not self._processDuringFlood(event):
+                msg = 'EVENT MANAGER: Not processing this event due to flood.'
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return
+            if self._prioritizeDuringFlood(event):
+                msg = 'EVENT MANAGER: Pruning event queue due to flood.'
+                debug.println(debug.LEVEL_INFO, msg, True)
+                self._pruneEventsDuringFlood()
+
+        if eType.startswith('object:selection-changed') \
+           and event.source in self._parentsOfDefunctDescendants:
+            msg = 'EVENT MANAGER: Ignoring event from parent of defunct descendants'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return
+
         if not debug.eventDebugFilter or debug.eventDebugFilter.match(eType) \
            and not eType.startswith("mouse:"):
-            debug.printDetails(debug.LEVEL_INFO, ' ' * 11, event.source)
+            indent = " " * 32
+            debug.printDetails(debug.LEVEL_INFO, indent, event.source)
+            if isinstance(event.any_data, pyatspi.Accessible):
+                debug.println(debug.LEVEL_INFO, '%sANY DATA:' % indent)
+                debug.printDetails(debug.LEVEL_INFO, indent, event.any_data, includeApp=False)
 
         script = self._getScriptForEvent(event)
         if not script:
@@ -752,6 +1063,29 @@ class EventManager:
             msg = 'EVENT MANAGER: %s: %s' % (key, value)
             debug.println(debug.LEVEL_INFO, msg, True)
 
+    def _processNewKeyboardEvent(self, device, pressed, keycode, keysym, state, text):
+        event = Atspi.DeviceEvent()
+        if pressed:
+            event.type = pyatspi.KEY_PRESSED_EVENT
+        else:
+            event.type = pyatspi.KEY_RELEASED_EVENT
+        event.hw_code = keycode
+        event.id = keysym
+        event.modifiers = state
+        event.event_string = text
+        if event.event_string is None:
+            event.event_string = ""
+        event.timestamp = device.event_count
+        device.event_count = device.event_count + 1
+
+        if not pressed and text == "Num_Lock" and "KP_Insert" in settings.orcaModifierKeys and orca_state.activeSWcript is not None:
+            orca_state.activeScript.refreshKeyGrabs()
+
+        if pressed:
+            orca_state.openingDialog = (text == "space" and (state & ~(1 << pyatspi.MODIFIER_NUMLOCK)))
+
+        self._processKeyboardEvent(event)
+
     def _processKeyboardEvent(self, event):
         keyboardEvent = input_event.KeyboardEvent(event)
         if not keyboardEvent.is_duplicate:
